@@ -21,15 +21,18 @@ class OpsAttrs:
     use_reports = None
     write_only = None
     cpus = None         #cpus list string
+    tids = None         #thread ids list string
 
-    def __init__(self, use_reports=False, write_only=False, cpus=None):
+    def __init__(self, use_reports=False, write_only=False, cpus=None,
+                 tids=None):
         self.use_reports = _damo_fmt_str.text_to_bool(use_reports)
         self.write_only = _damo_fmt_str.text_to_bool(write_only)
         self.cpus = cpus if cpus is not None else 'all'
+        self.tids = tids if tids is not None else ''
 
     def to_str(self, raw):
-        return 'use_reports: %s, write_only: %s, cpus: %s' % (
-                self.use_reports, self.write_only, self.cpus)
+        return 'use_reports: %s, write_only: %s, cpus: %s, threads: %s' % (
+                self.use_reports, self.write_only, self.cpus, self.tids)
 
     def __str__(self):
         return self.to_str(False)
@@ -38,7 +41,8 @@ class OpsAttrs:
         return type(self) == type(other) and \
                 self.use_reports == other.use_reports and \
                 self.write_only == other.write_only and \
-                self.cpus == other.cpus
+                self.cpus == other.cpus and \
+                self.tids == other.tids
 
     @classmethod
     def from_kvpairs(cls, kvpairs):
@@ -47,14 +51,16 @@ class OpsAttrs:
         write_only = kvpairs['write_only'] \
                 if 'write_only' in kvpairs else False
         cpus = kvpairs['cpus'] if 'cpus' in kvpairs else 'all'
+        tids = kvpairs['tids'] if 'tids' in kvpairs else ''
         return OpsAttrs(use_reports=use_reports, write_only=write_only,
-                        cpus=cpus)
+                        cpus=cpus, tids=tids)
 
     def to_kvpairs(self, raw=False):
         return collections.OrderedDict([
             ('use_reports', self.use_reports),
             ('write_only', self.write_only),
             ('cpus', self.cpus),
+            ('tids', self.tids),
             ])
 
 class DamonIntervalsGoal:
@@ -387,17 +393,21 @@ class DamonRegion:
 
 class DamonTarget:
     pid = None
+    obsolete = None
     regions = None
     context = None
 
-    def __init__(self, pid, regions):
+    def __init__(self, pid, regions, obsolete=False):
         self.pid = pid
         self.regions = regions
+        self.obsolete = _damo_fmt_str.text_to_bool(obsolete)
 
     def to_str(self, raw):
         lines = []
         if self.pid is not None:
             lines.append('pid: %s' % self.pid)
+        if self.obsolete is True:
+            line.sappend('(obsolete)')
         for region in self.regions:
             lines.append('region %s' % region.to_str(raw))
         return '\n'.join(lines)
@@ -411,11 +421,15 @@ class DamonTarget:
     @classmethod
     def from_kvpairs(cls, kvpairs):
         regions = [DamonRegion.from_kvpairs(kvp) for kvp in kvpairs['regions']]
-        return DamonTarget(kvpairs['pid'], regions)
+        obsolete = False
+        if 'obsolete' in kvpairs:
+            obsolete = kvpairs['obsolete']
+        return DamonTarget(kvpairs['pid'], regions, obsolete=obsolete)
 
     def to_kvpairs(self, raw=False):
         kvp = collections.OrderedDict()
         kvp['pid'] = self.pid
+        kvp['obsolete'] = self.obsolete
         kvp['regions'] = [r.to_kvpairs(raw) for r in self.regions]
         return kvp
 
@@ -567,10 +581,13 @@ qgoal_user_input = 'user_input'
 qgoal_some_mem_psi_us = 'some_mem_psi_us'
 qgoal_node_mem_used_bp = 'node_mem_used_bp'
 qgoal_node_mem_free_bp = 'node_mem_free_bp'
+qgoal_node_memcg_used_bp = 'node_memcg_used_bp'
+qgoal_node_memcg_free_bp = 'node_memcg_free_bp'
 qgoal_active_mem_bp = 'active_mem_bp'
 qgoal_inactive_mem_bp = 'inactive_mem_bp'
 qgoal_metrics = [qgoal_user_input, qgoal_some_mem_psi_us,
                  qgoal_node_mem_used_bp, qgoal_node_mem_free_bp,
+                 qgoal_node_memcg_used_bp, qgoal_node_memcg_free_bp,
                  qgoal_active_mem_bp, qgoal_inactive_mem_bp]
 
 class DamosQuotaGoal:
@@ -578,16 +595,19 @@ class DamosQuotaGoal:
     target_value = None
     current_value = None
     nid = None
+    memcg_path = None
     quotas = None
 
     def __init__(self, metric=qgoal_user_input,
-                 target_value='0', current_value='0', nid=None):
+                 target_value='0', current_value='0', nid=None,
+                 memcg_path=None):
         if not metric in qgoal_metrics:
             raise Exception('unsupported DAMOS quota goal metric')
         self.metric = metric
         if metric == qgoal_some_mem_psi_us:
             self.target_value = _damo_fmt_str.text_to_us(target_value)
         elif metric in [qgoal_node_mem_used_bp, qgoal_node_mem_free_bp,
+                        qgoal_node_memcg_used_bp, qgoal_node_memcg_free_bp,
                         qgoal_active_mem_bp, qgoal_inactive_mem_bp]:
             self.target_value = _damo_fmt_str.text_to_bp(target_value)
         else:
@@ -597,6 +617,7 @@ class DamosQuotaGoal:
             self.nid = None
         else:
             self.nid = _damo_fmt_str.text_to_nr(nid)
+        self.memcg_path = memcg_path
 
     @classmethod
     def metric_require_nid(cls, metric):
@@ -605,11 +626,24 @@ class DamosQuotaGoal:
     def has_nid(self):
         return DamosQuotaGoal.metric_require_nid(self.metric)
 
+    @classmethod
+    def metric_require_memcg_path(cls, metric):
+        return metric in [qgoal_node_memcg_used_bp, qgoal_node_memcg_free_bp]
+
+    def has_memcg_path(self):
+        return DamosQuotaGoal.metric_require_memcg_path(self.metric)
+
     def to_str(self, raw):
         metric_str = self.metric
+        additional_words = []
         if self.has_nid():
-            metric_str = '%s (nid %s)' % (
-                    metric_str, _damo_fmt_str.format_nr(self.nid, raw))
+            additional_words.append('nid %s' %
+                                    _damo_fmt_str.format_nr(self.nid, raw))
+        if self.has_memcg_path():
+            additional_words.append('memcg %s' % self.memcg_path)
+
+        if len(additional_words) > 0:
+            metric_str = '%s (%s)' % (metric_str, ', '.join(additional_words))
         return 'metric %s target %s current %s' % (
                 metric_str,
                 _damo_fmt_str.format_nr(self.target_value, raw),
@@ -621,6 +655,7 @@ class DamosQuotaGoal:
     def __eq__(self, other):
         return (type(self) == type(other) and self.metric == other.metric and
                 self.nid == other.nid and
+                self.memcg_path == other.memcg_path and
                 self.target_value == other.target_value and
                 self.current_value == other.current_value)
 
@@ -631,8 +666,13 @@ class DamosQuotaGoal:
             # later.
             return DamosQuotaGoal(target_value=kv['target_value_bp'],
                                   current_value=kv['current_value_bp'])
+        if 'memcg_path' in kv:
+            memcg_path = kv['memcg_path']
+        else:
+            memcg_path = None
         return DamosQuotaGoal(
                 metric=kv['metric'], nid=kv['nid'] if 'nid' in kv else None,
+                memcg_path=memcg_path,
                 target_value=kv['target_value'],
                 current_value=kv['current_value'])
 
@@ -641,6 +681,7 @@ class DamosQuotaGoal:
             ('metric', self.metric),
             ('nid', _damo_fmt_str.format_nr(self.nid, raw)
              if self.nid is not None else None),
+            ('memcg_path', self.memcg_path),
             ('target_value', _damo_fmt_str.format_nr(self.target_value,
                 raw)),
             ('current_value', _damo_fmt_str.format_nr(self.current_value,
@@ -1047,7 +1088,7 @@ class Damos:
             if self.apply_interval_us != 0 else 'per aggr interval')
         return ' '.join(action_words)
 
-    def to_str(self, raw, params_only=False):
+    def to_str(self, raw, params_only=False, omit_tried_regions=False):
         lines = [self.str_action_line(raw)]
         if self.access_pattern is not None:
             lines.append('target access pattern')
@@ -1070,7 +1111,8 @@ class Damos:
         if params_only is False and self.stats is not None:
             lines.append('statistics')
             lines.append(_damo_fmt_str.indent_lines(self.stats.to_str(raw), 4))
-        if params_only is False and self.tried_regions is not None:
+        if params_only is False and omit_tried_regions is False and \
+                self.tried_regions is not None:
             lines.append('tried regions (%s)' % _damo_fmt_str.format_sz(
                     self.tried_bytes, raw))
             for region in self.tried_regions:
@@ -1172,11 +1214,12 @@ class DamonCtx:
         for scheme in self.schemes:
             scheme.context = self
 
-    def to_str(self, raw, params_only=False):
+    def to_str(self, raw, params_only=False, omit_damos_tried_regions=False):
         ops_line_tokens = ['ops: %s' % self.ops]
         if self.ops_attrs.use_reports:
             ops_line_tokens.append('use_reports')
             ops_line_tokens.append('cpus %s' % self.ops_attrs.cpus)
+            ops_line_tokens.append('threads %s' % self.ops_attrs.tids)
         if self.ops_attrs.write_only:
             ops_line_tokens.append('write-only')
         lines = [' '.join(ops_line_tokens)]
@@ -1194,7 +1237,7 @@ class DamonCtx:
         for idx, scheme in enumerate(self.schemes):
             lines.append('scheme %d' % idx)
             lines.append(_damo_fmt_str.indent_lines(
-                scheme.to_str(raw, params_only), 4))
+                scheme.to_str(raw, params_only, omit_damos_tried_regions), 4))
         return '\n'.join(lines)
 
     def __str__(self):
@@ -1263,7 +1306,8 @@ class Kdamond:
             words.append('cpu usage: %s' % self.get_cpu_usage())
         return ', '.join(words)
 
-    def to_str(self, raw, show_cpu=False, params_only=False):
+    def to_str(self, raw, show_cpu=False, params_only=False,
+               omit_damos_tried_regions=False):
         lines = []
         summary_line = self.summary_str(show_cpu, params_only,
                                         omit_defaults=False, raw_number=raw)
@@ -1272,7 +1316,7 @@ class Kdamond:
         for idx, ctx in enumerate(self.contexts):
             lines.append('context %d' % idx)
             lines.append(_damo_fmt_str.indent_lines(
-                ctx.to_str(raw, params_only), 4))
+                ctx.to_str(raw, params_only, omit_damos_tried_regions), 4))
         return '\n'.join(lines)
 
     def __str__(self):
@@ -1383,6 +1427,9 @@ features = ['record',       # was in DAMON patchset, but not merged in mainline
                                         # merged in v6.16-rc1
             'schemes_dests',            # merged in v6.17-rc1
             'sysfs_refresh_ms',         # merged in v6.17-rc1
+            'addr_unit',                # merged in v6.18-rc1
+            'schemes_quota_goal_node_memcg_used_free',
+                                        # merged in mm, expectred for 6.19-rc1
             'ops_attrs',                # hacking on damon/next
             ]
 
@@ -1663,6 +1710,82 @@ def update_schemes_status(stats=True, tried_regions=True,
 
     return None
 
+def get_childs_pids(pid):
+    try:
+        childs_pids = subprocess.check_output(
+                ['ps', '--ppid', pid, '-o', 'pid=']
+                ).decode().split()
+    except:
+        childs_pids = []
+
+    ret = childs_pids
+    for child_pid in childs_pids:
+        childs_childs_pids = get_childs_pids(child_pid)
+        ret.extend(childs_childs_pids)
+
+    return ret
+
+def pid_running(pid):
+    '''pid should be string'''
+    try:
+        subprocess.check_output(['ps', '--pid', pid])
+        return True
+    except:
+        return False
+
+def add_childs_target(kdamonds):
+    # TODO: Support multiple kdamonds
+    if not target_has_pid(kdamonds[0].contexts[0].ops):
+        return
+    old_targets = kdamonds[0].contexts[0].targets
+    current_target_pids = []
+    new_target_pids = []
+
+    for target in old_targets:
+        if target is None:
+            continue
+        current_target_pids.append('%s' % target.pid)
+
+    for target_pid in current_target_pids:
+        # If this is already in new_target_pids, we have already recursively
+        # searched for its children as well.
+        if target_pid in new_target_pids:
+            continue
+        # Don't consider already terminated targets since committing already
+        # terminated targets to DAMON fails
+        if not pid_running(target_pid):
+            continue
+
+        new_target_pids.append(target_pid)
+        childs_pids = get_childs_pids(target_pid)
+        if len(childs_pids) == 0:
+            continue
+
+        for child_pid in childs_pids:
+            # skip the child if already in the targets
+            if child_pid in new_target_pids:
+                continue
+            new_target_pids.append(child_pid)
+
+    if set(new_target_pids) == set(current_target_pids):
+        return None
+    # Commit the new set of targets.
+    #
+    # Set no target region because it will instruct DAMON to keep current
+    # monitoring results (regions with nr_accesses, age, etc) for the target if
+    # exists.
+    #
+    # TODO: adjust the orders of the new targets so that only appropriate past
+    # monitoring results are inherited.
+    kdamonds[0].contexts[0].targets = [DamonTarget(pid = p, regions=[])
+                                       for p in new_target_pids]
+    err = commit(kdamonds, commit_targets_only=True)
+    if err is not None:
+        kdamonds[0].contexts[0].targets = old_targets
+        return 'commit failed (%s)' % err
+
+    return None
+
 def turn_damon_on(kdamonds_idxs):
     err = _damon_fs.turn_damon_on(kdamonds_idxs)
     if err:
