@@ -260,18 +260,29 @@ def damos_quotas_cons_arg(cmd_args):
 
 def damos_options_to_quota_goal(garg):
     # garg is the user inputs
-    # garg should be <metric> <target value> [<current value>|<nid>]
-    # [current value] is given for only 'user_input' <metric>
-    # [nid] is given for only node_mem_{used,free}_bp
-    if not len(garg) in [2, 3]:
+    # garg should be <metric> <target value> [<optional>...]
+    # for user_input, one optional argument for "current value" is given.
+    # for node_mem[cg]_{used,free}_bp, one optional argument for node id is
+    # given.
+    # for node_memcg_{used,free}_bp, one more optional argument for memcg path
+    # is given.
+    if not len(garg) in [2, 3, 4]:
         return None, 'Wrong --damos_quota_goal (%s)' % garg
     metric, target_value, optionals = garg[0], garg[1], garg[2:]
     current_value = 0
     nid = None
-    if _damon.DamosQuotaGoal.metric_require_nid(metric):
+    memcg_path = None
+    if metric in [_damon.qgoal_node_mem_used_bp,
+                  _damon.qgoal_node_mem_free_bp]:
         if len(optionals) != 1:
             return None, 'nid is not given or something else is given'
         nid = optionals[0]
+    elif metric in [_damon.qgoal_node_memcg_used_bp,
+                    _damon.qgoal_node_memcg_free_bp]:
+        if len(optionals) != 2:
+            return None, 'nid and memcg_path required'
+        nid = optionals[0]
+        memcg_path = optionals[1]
     elif metric == 'user_input':
         if len(optionals) != 1:
             return None, 'current value is not given or something else is given'
@@ -279,7 +290,8 @@ def damos_options_to_quota_goal(garg):
     try:
         return _damon.DamosQuotaGoal(
                 metric=metric, target_value=target_value,
-                current_value=current_value, nid=nid), None
+                current_value=current_value,
+                nid=nid, memcg_path=memcg_path), None
     except Exception as e:
         return None, 'DamosQuotaGoal creation fail (%s, %s)' % (garg, e)
 
@@ -331,47 +343,66 @@ def damos_options_to_scheme(sz_region, access_rate, age, action,
     except Exception as e:
         return None, 'Wrong \'--damos_*\' argument (%s)' % e
 
-def damos_options_to_schemes(args):
-    if args.damos_quota_interval:
-        for i, interval in enumerate(args.damos_quota_interval):
-            t, s = 0, 0
-            if i < len(args.damos_quota_time):
-                t = args.damos_quota_time[i]
-            if i < len(args.damos_quota_space):
-                s = args.damos_quota_space[i]
-            if i < len(args.damos_quota_weights):
-                w1, w2, w3 = args.damos_quota_weights[i]
-            else:
-                w1, w2, w3 = 1, 1, 1
-            args.damos_quotas.append([t, s, interval, w1, w2, w3])
+def set_args_damos_quotas(args):
+    # DAMOS quotas can be set using
+    # 1) only --damos_quotas (specify time, space, interval, weights at once),
+    #    or
+    # 2) --damos_quota_interval, --damos_quota_time, --damos_quota_space,
+    #    and --damos_quota_weights (specify the components one by one).
+    # Convert quotas specified in 2) way into --damos_quotas for easier
+    # handling.
+    if not args.damos_quota_interval:
+        return
+    for i, interval in enumerate(args.damos_quota_interval):
+        t, s = 0, 0
+        if i < len(args.damos_quota_time):
+            t = args.damos_quota_time[i]
+        if i < len(args.damos_quota_space):
+            s = args.damos_quota_space[i]
+        if i < len(args.damos_quota_weights):
+            w1, w2, w3 = args.damos_quota_weights[i]
+        else:
+            w1, w2, w3 = 1, 1, 1
+        args.damos_quotas.append([t, s, interval, w1, w2, w3])
+
+def verify_set_damos_args_len(args):
+    # return an error string if given args is wrong in terms of lengths.
     nr_schemes = len(args.damos_action)
+
     if len(args.damos_sz_region) > nr_schemes:
-        return [], 'too much --damos_sz_region'
+        return 'too much --damos_sz_region'
     if len(args.damos_access_rate) > nr_schemes:
-        return [], 'too much --damos_access_rate'
+        return 'too much --damos_access_rate'
     if len(args.damos_age) > nr_schemes:
-        return [], 'too much --damos_age'
+        return 'too much --damos_age'
     if len(args.damos_apply_interval) > nr_schemes:
-        return [], 'too much --damos_apply_interval'
+        return 'too much --damos_apply_interval'
     if len(args.damos_quotas) > nr_schemes:
-        return [], 'too much --damos_quotas'
+        return 'too much --damos_quotas'
+
     if len(args.damos_quota_goal) > 0 and nr_schemes > 1:
         if len(args.damos_nr_quota_goals) == 0:
-            return [], '--damos_nr_quota_goals required'
+            return '--damos_nr_quota_goals required'
     if nr_schemes == 1 and args.damos_nr_quota_goals == []:
         args.damos_nr_quota_goals = [len(args.damos_quota_goal)]
     if sum(args.damos_nr_quota_goals) != len(args.damos_quota_goal):
-        return [], 'wrong --damos_nr_quota_goals'
+        return 'wrong --damos_nr_quota_goals'
+
     if len(args.damos_wmarks) > nr_schemes:
-        return [], 'too much --damos_wmarks'
+        return 'too much --damos_wmarks'
+
     # for multiple schemes, number of filters per scheme is required
     if len(args.damos_filter) > 0 and nr_schemes > 1:
         if len(args.damos_nr_filters) == 0:
-            return [], '--damos_nr_filters required'
+            return '--damos_nr_filters required'
     if nr_schemes == 1 and args.damos_nr_filters == []:
         args.damos_nr_filters = [len(args.damos_filter)]
     if sum(args.damos_nr_filters) != len(args.damos_filter):
-        return [], 'wrong --damos_nr_filters'
+        return 'wrong --damos_nr_filters'
+    return None
+
+def fillup_default_damos_args(args):
+    nr_schemes = len(args.damos_action)
 
     args.damos_sz_region += [['min', 'max']] * (
             nr_schemes - len(args.damos_sz_region))
@@ -382,33 +413,38 @@ def damos_options_to_schemes(args):
             nr_schemes - len(args.damos_apply_interval))
     args.damos_quotas += [None] * (nr_schemes - len(args.damos_quotas))
     args.damos_wmarks += [None] * (nr_schemes - len(args.damos_wmarks))
-    target_nid = [None] * nr_schemes
-    dests_list = []
-    for i in range(nr_schemes):
-        dests_list.append([])
+
+def damos_options_to_schemes(args):
+    set_args_damos_quotas(args)
+    err = verify_set_damos_args_len(args)
+    if err is not None:
+        return [], err
+    fillup_default_damos_args(args)
+    nr_schemes = len(args.damos_action)
+
     schemes = []
 
     for i in range(nr_schemes):
         action = args.damos_action[i][0]
+        target_nid = None
+        dests = []
         if _damon.is_damos_migrate_action(action):
             try:
                 if len(args.damos_action[i]) == 2:
-                    target_nid[i] = int(args.damos_action[i][1])
-                    args.damos_action[i] = args.damos_action[i][0]
+                    target_nid = int(args.damos_action[i][1])
                 elif len(args.damos_action[i]) > 2:
                     dests = []
                     for j in range(1, len(args.damos_action[i]), 2):
                         dests.append(_damon.DamosDest(
                             args.damos_action[i][j],
                             args.damos_action[i][j + 1]))
-                    dests_list[i] = dests
-                    args.damos_action[i] = args.damos_action[i][0]
             except:
                 return [], '"%s" action require a numeric target_nid ' \
                         'or dests arguments.' \
                             % args.damos_action[i][0]
-        else:
-            args.damos_action[i] = action
+        elif len(args.damos_action[i]) > 1:
+            return [], 'Wrong number of --damos_action arguments.' % action
+        args.damos_action[i] = action
 
         qgoals = []
         if args.damos_quota_goal:
@@ -426,7 +462,7 @@ def damos_options_to_schemes(args):
                 args.damos_access_rate[i], args.damos_age[i],
                 args.damos_action[i], args.damos_apply_interval[i],
                 args.damos_quotas[i], qgoals, args.damos_wmarks[i],
-                target_nid[i], filters, dests_list[i])
+                target_nid, filters, dests)
         if err != None:
             return [], err
         schemes.append(scheme)
@@ -493,9 +529,14 @@ def damon_ctx_for(args, idx):
         cpus = args.exp_ops_cpus[idx]
     else:
         cpus = 'all'
+    if args.exp_ops_tids[idx] is not None:
+        tids = args.exp_ops_tids[idx]
+    else:
+        tids = ''
     try:
         ops_attrs = _damon.OpsAttrs(
-                use_reports=use_reports, write_only=write_only, cpus=cpus)
+                use_reports=use_reports, write_only=write_only, cpus=cpus,
+                tids=tids)
     except Exception as e:
         return None, 'invalid ops_attrs arguments (%s)' % e
 
@@ -534,8 +575,9 @@ def fillup_none_ctx_args(args):
     nr_ctxs = get_nr_ctxs(args)
     for attr_name in [
             'ops', 'exp_ops_use_reports', 'exp_ops_write_only', 'exp_ops_cpus',
-            'sample', 'aggr', 'updr', 'minr', 'maxr', 'monitoring_intervals',
-            'monitoring_intervals_goal', 'monitoring_nr_regions_range']:
+            'exp_ops_tids', 'sample', 'aggr', 'updr', 'minr', 'maxr',
+            'monitoring_intervals', 'monitoring_intervals_goal',
+            'monitoring_nr_regions_range']:
         attr_val = getattr(args, attr_name)
         if attr_val is None:
             setattr(args, attr_name, [None] * nr_ctxs)
@@ -882,6 +924,9 @@ def set_monitoring_damos_common_args(parser, hide_help=False):
     parser.add_argument('--exp_ops_cpus', action='append', metavar='<cpulist>',
                         help='monitor for given cpus only (experimental)'
                         if not hide_help else argparse.SUPPRESS)
+    parser.add_argument(
+            '--exp_ops_tids', action='append', metavar='<thread ids list>',
+            help='monitoring for given threads only (experimental)')
     parser.add_argument(
             '--refresh_stat', metavar='<milliseconds>', action='append',
             help='automatic kdamond internal stat refresh interval')
