@@ -282,6 +282,12 @@ def write_quota_goal_dir(dir_path, goal):
         if err is not None:
             return err
 
+    if goal.has_memcg_path():
+        err = _damo_fs.write_file(
+                os.path.join(dir_path, 'path'), '%s' % goal.memcg_path)
+        if err is not None:
+            return err
+
     err = _damo_fs.write_file(
             os.path.join(dir_path, 'target_value'),
             '%d' % goal.target_value)
@@ -477,6 +483,15 @@ def write_target_dir(dir_path, target):
         if err is not None:
             return err
 
+    obsolete_file = os.path.join(dir_path, 'obsolete_target')
+    if os.path.isfile(obsolete_file):
+        err = _damo_fs.write_file(
+                obsolete_file, '1' if target.obsolete else '0')
+        if err is not None:
+            return err
+    elif target.obsolete:
+        return 'obsolete_target unsupported'
+
     return write_target_regions_dir(
             os.path.join(dir_path, 'regions'), target.regions)
 
@@ -521,7 +536,12 @@ def write_ops_attrs_dir(dir_path, ops_attrs):
                                'Y' if ops_attrs.write_only else 'N')
     if err is not None:
         return err
-    return _damo_fs.write_file(os.path.join(dir_path, 'cpus'), ops_attrs.cpus)
+    err = _damo_fs.write_file(os.path.join(dir_path, 'cpus'), ops_attrs.cpus)
+    if err is not None:
+        return err
+    err = _damo_fs.write_file(os.path.join(dir_path, 'tids'), ops_attrs.tids)
+    if err is not None:
+        return err
 
 def write_monitoring_attrs_dir(dir_path, context):
     err = _damo_fs.write_file(
@@ -692,10 +712,15 @@ def files_content_to_quota_goals(files_content):
     goals = []
     for goal_kv in number_sorted_dirs(files_content):
         if 'target_metric' in goal_kv:
+            if 'path' in goal_kv:
+                memcg_path = goal_kv['path'].strip()
+            else:
+                memcg_path = None
             goals.append(
                     _damon.DamosQuotaGoal(
                         metric=goal_kv['target_metric'].strip(),
                         nid=goal_kv['nid'] if 'nid' in goal_kv else None,
+                        memcg_path=memcg_path,
                         target_value=goal_kv['target_value'],
                         current_value=goal_kv['current_value']))
         else:
@@ -825,15 +850,19 @@ def files_content_to_target(files_content):
         pid = int(files_content['pid_target'])
     except:
         pid = None
+    obsolete = False
+    if 'obsolete_target' in files_content:
+        obsolete = files_content['obsolete_target'].strip()
     regions = files_content_to_regions(files_content['regions'])
-    return _damon.DamonTarget(pid, regions)
+    return _damon.DamonTarget(pid, regions, obsolete=obsolete)
 
 def files_content_to_ops_attrs(files_content):
     use_reports = files_content['use_reports'].strip()
     write_only = files_content['write_only'].strip()
     cpus = files_content['cpus'].strip()
+    tids = files_content['tids'].strip()
     return _damon.OpsAttrs(use_reports=use_reports, write_only=write_only,
-                           cpus=cpus)
+                           cpus=cpus, tids=tids)
 
 def files_content_to_context(files_content):
     mon_attrs_content = files_content['monitoring_attrs']
@@ -994,7 +1023,9 @@ def infer_damon_version():
                                 )])])]
     err = stage_kdamonds(kdamonds)
     if err is None:
-        if os.path.isfile(os.path.join(kdamond_dir_of(0), 'refresh_ms')):
+        if os.path.isfile(os.path.join(ctx_dir_of(0, 0), 'addr_unit')):
+            version = '>v6.17'
+        elif os.path.isfile(os.path.join(kdamond_dir_of(0), 'refresh_ms')):
             version = '>v6.16'
         elif os.path.isdir(os.path.join(scheme_dir_of(0, 0, 0), 'dests')):
             version = '>v6.16'
@@ -1157,12 +1188,20 @@ def update_supported_features():
                              'nid')):
             feature_supports['schemes_quota_goal_node_mem_used_free'] = True
 
+        if os.path.isfile(
+                os.path.join(scheme_dir_of(0, 0, 0), 'quotas', 'goals', '0',
+                             'path')):
+            feature_supports['schemes_quota_goal_node_memcg_used_free'] = True
+
     if os.path.isdir(os.path.join(scheme_dir_of(0, 0, 0), 'dests')):
         feature_supports['schemes_dests'] = True
 
     if os.path.isfile(os.path.join(kdamond_dir_of(0), 'refresh_ms')):
         feature_supports['sysfs_refresh_ms'] = True
 
+    if os.path.isfile(os.path.join(ctx_dir_of(0, 0), 'addr_unit')):
+        feature_supports['addr_unit'] = True
+
     if os.path.isdir(os.path.join(ctx_dir_of(0, 0), 'operations_attrs')):
         feature_supports['ops_attrs'] = True
 
